import os, cytoplasm, sys

# make this work in either Python 2.x or 3.x
if sys.version_info.major >= 3:
    from http.server import SimpleHTTPRequestHandler
else:
    from SimpleHTTPServer import SimpleHTTPRequestHandler

# keep track of when things were last built in this global variable
most_recent_time = 0

def handler(rebuild):
    # build once first...
    cytoplasm.build()
    # and save the most recent time
    global most_recent_time
    most_recent_time = most_recent(".")
    # change to the build directory, where things are to be served from.
    os.chdir(cytoplasm.configuration.build_dir)
    if rebuild:
        # if rebuild is true, return the RebuildHandler
        return RebuildHandler
    else:
        # otherwise, just return the SimpleHTTPRequestHandler
        return SimpleHTTPRequestHandler

def most_recent(directory):
    "Determine the most recent modified time in the source directory, ignoring dotfiles and _build."
    build_dir = cytoplasm.configuration.build_dir
    # get the candidate files:
    files = [f for f in os.listdir(directory) if f != build_dir and not f.startswith(".")]
    # get each of their times
    times = [os.stat(os.path.join(directory, f)).st_mtime for f in files]
    # the highest time here is the most recent; return that.
    return max(times)


class RebuildHandler(SimpleHTTPRequestHandler):
    def do_GET(self):
        "Send a GET request, and, if anything has changed, rebuild the site."
        # overwrite the do_GET method in SimpleHTTPRequestHandler with this.
        # it's mostly the same except for the 'build' part here.
        # declare most_recent_time global; we'll be changing it later.
        global most_recent_time
        f = self.send_head()
        if f:
            # figure out what the most recent time edited is in the source directory
            new_recent = most_recent("..")
            # only build the site if the new most recent is more recent than the old one,
            # i.e. if one or more of the files has been edited.
            if new_recent > most_recent_time:
                # update most_recent_time
                most_recent_time = new_recent
                # The following is an ugly hack:
                # Move up one directory to the source directory
                os.chdir("..")
                # Build the site from the source directory
                cytoplasm.build()
                # and move back down to the build directory to continue serving.
                os.chdir(cytoplasm.configuration.build_dir)
            # Copy the file to self.wfile...
            self.copyfile(f, self.wfile)
            # and then close it.
            f.close()





from .errors import InterpreterError

def SaveReturned(fn):
    '''Some potential interpreters, like Mako, don't give you an easy way to save to a destination.
    In these cases, simply use this function as a decorater.'''
    def InterpreterWithSave(source, destination, **kwargs):
        # under some circumstances, this should be able to write to file-like objects;
        # so if destination is a string, assume it's a path; otherwise, it's a file-like object
        if isinstance(destination, str):
            f = open(destination, "w")
        else:
            f = destination
        # pass **kwargs to this function.
        f.write(fn(source, **kwargs))
        f.close()
    return InterpreterWithSave

def interpret(file, destination, **kwargs):
    "Interpret a file with an interpreter according to its suffix."
    from .configuration import interpreters
    for ending in interpreters.keys():
        # if the file has a suffix that matches any of the interpreters,
        # parse that file with that interpreter
        if file.endswith(".%s" %(ending)):
            # give the interpreter two variables -- the original place and hte destination
            # of the interpreted file.
            try:
                interpreters[ending](file, destination, **kwargs)
            except Exception as exception:
                # if the interpreter chokes, raise an InterpreterError with some useful information.
                raise InterpreterError("%s on file '%s': %s" %(ending, file, exception))
            handled = True
            return True
    return False




# This module contains the user's configurations, to be accessed like:
# `print cytoplasm.configuration.build_dir`
import os, imp
from .errors import CytoplasmError

# If the user has a file called _config.py, import that.
# The user's _config.py should "from cytoplasm.defaults import *" if they want to use
# some of the defaults.
if os.path.exists("_config.py"):
    imp.load_source("_config", "_config.py")
    from _config import *
else:
    raise CytoplasmError("You don't seem to have a configuration file at '_config.py'.")



import os, shutil
from . import errors, interpreters, configuration
from .controllers import controllerclass

def copy_over():
    "Copy the files not beginning with '_' to the site directory"
    interpreter_dictionary = configuration.interpreters
    # All of the files and directories that don't begin with '_' or '.'
    to_copy = [file for file in os.listdir(".") if not file.startswith(("_", "."))]
    files = [file for file in to_copy if not os.path.isdir(file)]
    for file in files:
        # first, try to interpret each of these files with one of the interpreters
        # if none of the interpreters match, it returns False
        # the destination, if interpreted, should be the filename without that last suffix
        destination = file.rsplit('.', 1)[0]
        if not interpreters.interpret(file, "%s/%s" %(configuration.build_dir, destination)):
            # otherwise, simply copy the file over
            shutil.copy2(file, configuration.build_dir)
    directories = [file for file in to_copy if os.path.isdir(file)]
    for dir in directories:
        if os.path.exists("%s/%s" %(configuration.build_dir, dir)):
            # if the directory exists, delete it.
            shutil.rmtree("%s/%s" %(configuration.build_dir, dir))
        # copy all the directories that don't start with "." or "_" completely.
        shutil.copytree(dir, "%s/%s" %(configuration.build_dir, dir))

def build():
    # Create the build directory, if it doesn't exist
    if not os.path.exists(configuration.build_dir):
        os.mkdir(configuration.build_dir)
    copy_over()
    for controller, arguments in configuration.controllers:
        # create a controller object of the class returned by controllerclass
        controller_object = controllerclass(controller)(*arguments)
        # call the controller object 
        controller_object()
        

